use proc_macro::TokenStream;
use quote::quote;
use std::ops::Deref;
use syn::__private::Span;
use syn::punctuated::Punctuated;
use syn::token::{Colon, PathSep, Plus};
use syn::{
    parse_macro_input, Data, DeriveInput, Expr, GenericParam, Ident, Lit, Meta, Path, PathSegment,
    Token, TraitBound, TraitBoundModifier, TypeParam, TypeParamBound,
};

// Struct name: Ads1256PollStateful, Ads1256PollPublish, Ads1256PollStatePub
#[proc_macro_derive(PollVariants, attributes(value_type))]
pub fn poll_variant_macro(input: TokenStream) -> TokenStream {
    // ----- Parse input ----------------------------------
    let input = parse_macro_input!(input as DeriveInput);
    let attrs = &input.attrs;
    let vis = &input.vis;
    let ident = &input.ident;
    let data = &input.data;
    let og_generics = &input.generics;
    let (og_impl_generics, og_type_generics, og_where_clause) = &og_generics.split_for_impl();

    // ----- Check that item the macro was used on is a struct ----------------------------------
    match data {
        Data::Struct(struct_data) => struct_data,
        Data::Enum(_) => panic!("Stateful struct cannot be derived from an enum."),
        Data::Union(_) => panic!("Stateful struct cannot be derived from a union."),
    };

    // ----- Extract value type attribute ----------------------------------
    const VALUE_T_NAME: &str = "value_type";
    let mut value_type_ident: Option<Ident> = None;
    for attribute in attrs.iter() {
        // if the attribute is a named value
        if let Meta::NameValue(meta) = &attribute.meta {
            // if the name of the attribute is value_t
            if meta.path.segments[0].ident == VALUE_T_NAME {
                // if the value of the attribute is a literal
                if let Expr::Lit(lit) = &meta.value {
                    // if the literal is a string
                    if let Lit::Str(lit) = &lit.lit {
                        value_type_ident = Some(Ident::new(lit.value().deref(), Span::call_site()));
                    } else {
                        panic!("{VALUE_T_NAME} must be set with a string literal.")
                    }
                } else {
                    panic!("{VALUE_T_NAME} must be set with a literal.")
                }
            } else {
                continue;
            }
        } else {
            continue;
        }
    }
    let value_type_ident = if let Some(val) = value_type_ident {
        val
    } else {
        panic!("Need attribute {VALUE_T_NAME}: #[{VALUE_T_NAME} = \"type\"]")
    };

    // ----- Build publisher generics ----------------------------------
    // Check if generics contains T
    const MUTEX_T_IDENT_STR: &str = "MutexT";
    let mut num_lifetimes: usize = 0;
    let mut num_types: usize = 0;
    let mut num_const: usize = 0;
    let mut has_mutex_t = false;
    for param in og_generics.params.iter() {
        match param {
            GenericParam::Lifetime(_) => num_lifetimes += 1,
            GenericParam::Type(param) => {
                num_types += 1;

                // If the generic parameter is T
                if param.ident == MUTEX_T_IDENT_STR {
                    has_mutex_t = true;
                }
            },
            GenericParam::Const(_) => num_const += 1,
        }
    }

    let mut publish_generics = og_generics.clone();
    // If MutexT is not a generic parameter, add it
    if !has_mutex_t {
        let mutex_t_ident = Ident::new(MUTEX_T_IDENT_STR, Span::call_site());
        let mut mutex_t_bounds: Punctuated<TypeParamBound, Plus> = Punctuated::new();

        let mut path: Punctuated<PathSegment, PathSep> = Punctuated::new();
        path.push(Ident::new("embassy_sync", Span::call_site()).into());
        path.push(Ident::new("blocking_mutex", Span::call_site()).into());
        path.push(Ident::new("raw", Span::call_site()).into());
        path.push(Ident::new("RawMutex", Span::call_site()).into());
        let raw_mutex_path = Path {
            leading_colon: None,
            segments: path,
        };

        let mutex_t_bound = TraitBound {
            paren_token: None,
            modifier: TraitBoundModifier::None,
            lifetimes: None,
            path: raw_mutex_path,
        };
        mutex_t_bounds.push(mutex_t_bound.into());

        let mutex_t_param: TypeParam = TypeParam {
            attrs: Vec::new(),
            ident: mutex_t_ident,
            colon_token: Some(Colon::default()),
            bounds: mutex_t_bounds,
            eq_token: None,
            default: None,
        };

        let num_generics = num_lifetimes + num_types + num_const;
        // If there are generics
        if num_generics > 0 {
            // If all generics are lifetimes
            if num_lifetimes == num_generics {
                // Add MutexT after the lifetimes
                publish_generics.params.push(mutex_t_param.into());
                // If no generics are lifetimes
            } else if num_lifetimes == 0 {
                // Insert MutexT at the front
                publish_generics
                    .params
                    .insert(0, mutex_t_param.into());
                // If there are lifetimes and other generics
            } else {
                // Insert MutexT after the lifetimes
                publish_generics
                    .params
                    .insert(num_lifetimes, mutex_t_param.into());
            }
            // If there are no generics
        } else {
            // Add MutexT
            publish_generics.params.push(mutex_t_param.into());
        }
    }
    // const generics


    let (publ_impl_generics, publ_type_generics, publ_where_clause) =
        &publish_generics.split_for_impl();

    let stateful_ident = Ident::new(format!("{ident}Stateful").deref(), ident.span());
    let publish_ident = Ident::new(format!("{ident}Publish").deref(), ident.span());
    let state_pub_ident = Ident::new(format!("{ident}StatePub").deref(), ident.span());

    let expanded = quote! {
        // ----- Stateful struct ----------------------------------
        #vis struct #stateful_ident #og_generics #og_where_clause {
            pub poll: #ident #og_type_generics,
            pub state: physical_node::transducer::State<#value_type_ident>,
        }

        // ----- Stateful impls ----------------------------------
        impl #og_impl_generics physical_node::transducer::Stateful for #stateful_ident #og_type_generics #og_where_clause {
            type Value = #value_type_ident;

            #[inline(always)]
            fn state_cell(&self) -> physical_node::cell::CellView<Self::Value> {
                self.state.state_cell()
            }

            #[inline(always)]
            fn state(&self) -> Self::Value {
                self.state.state()
            }
        }

        impl #og_impl_generics physical_node::transducer::input::Poll for #stateful_ident #og_type_generics #og_where_clause {
            type Value = #value_type_ident;

            #[inline(always)]
            async fn poll(&self) -> Self::Value {
                let value = self.poll.poll().await;
                self.state.update(value);
                value
            }
        }

        // ----- Publish struct ----------------------------------
        #[cfg(feature = "embassy-sync")]
        #vis struct #publish_ident #publish_generics #publ_where_clause {
            pub poll: #ident #og_type_generics,
            pub publisher: physical_node::transducer::Publisher<#value_type_ident, MutexT, CAPACITY, NUM_SUBS>,
        }

        // ----- Publish impl ----------------------------------

        // ----- StatePub struct ----------------------------------

        // ----- StatePub impl ----------------------------------
    };

    TokenStream::from(expanded)
}
